static_template = """
# 📃 Role overview

## {{ role.name }}

{% if role.belongs_to_collection -%}
```
Role belongs to {{ role.belongs_to_collection.namespace }}/{{ role.belongs_to_collection.name }}
Namespace - {{ role.belongs_to_collection.namespace }}
Collection - {{ role.belongs_to_collection.name }}
Version - {{ role.belongs_to_collection.version }}
Repository - {{ role.belongs_to_collection.repository }}
```
{%- endif %}

{% if role.meta and role.meta.galaxy_info -%}
Description: {{ role.meta.galaxy_info.description or 'Not available.' }}
{% else %}
Description: Not available.
{%- endif %}

{% if role.docsible -%}
| Field                | Value           |
|--------------------- |-----------------|
{%- if role.docsible.description %}
| Functional description | {{ role.docsible.description }} |
{%- endif %}
{%- if role.docsible.requester %}
| Requester            | {{ role.docsible.requester }} |
{%- endif %}
{%- if role.docsible.users %}
| Users                | {{ role.docsible.users }} |
{%- endif %}
{%- if role.docsible.dt_dev %}
| Date dev             | {{ role.docsible.dt_dev }} |
{%- endif %}
{%- if role.docsible.dt_prod %}
| Date prod            | {{ role.docsible.dt_prod }} |
{%- endif %}
{%- if role.docsible.dt_update %}
| Readme update        | {{ role.docsible.dt_update }} |
{%- endif %}
{%- if role.docsible.version %}
| Version              | {{ role.docsible.version }} |
{%- endif %}
{%- if role.docsible.time_saving %}
| Time Saving          | {{ role.docsible.time_saving }} |
{%- endif %}
{%- if role.docsible.category %}
| Category             | {{ role.docsible.category }} |
{%- endif %}
{%- if role.docsible.subCategory %}
| Sub category         | {{ role.docsible.subCategory }} |
{%- endif %}
{%- if role.docsible.aap_hub %}
| AAP Hub              | {{ role.docsible.aap_hub }} |
{%- endif %}
{%- if role.docsible.automation_kind %}
| Automation Kind      | {{ role.docsible.automation_kind }} |
{%- endif %}
{%- if role.docsible.critical %}
| Critical ⚠️          | {{ role.docsible.critical }} |
{%- endif %}
{%- endif %}

{% macro render_arguments_list(arguments, level=0) %}
{% for arg, details in arguments.items() %}
  {%- set indent = '  ' * level %}
  {{ indent }}- **{{ arg }}**
  {{ indent }}  - **Required**: {{ details.required | default('false') }}
  {{ indent }}  - **Type**: {{ details.type }}
  {{ indent }}  - **Default**: {{ details.default | default('none') }}
  {{ indent }}  - **Description**: {{ details.description | default('No description provided') }}
  {% if details.choices is defined %}
    {{ indent }}  - **Choices**:
    {% for choice in details.choices %}
      {{ indent }}    - {{ choice }}
    {% endfor %}
  {% endif %}
  {% if details.aliases is defined %}
    {{ indent }}  - **Aliases**:
    {% for alias in details.aliases %}
      {{ indent }}    - {{ alias }}
    {% endfor %}
  {% endif %}
  {% if details.type == 'dict' and details.options %}
    {{ render_arguments_list(details.options, level + 1) }}
  {% elif details.type == 'list' and details.elements == 'dict' %}
    {% for elem in details.default %}
      {% if elem is mapping %}
        {{ render_arguments_list(elem, level + 1) }}
      {% endif %}
    {% endfor %}
  {% endif %}
{% endfor %}
{% endmacro %}

{% if role.argument_specs %}
<details>
<summary><b>🧩 Argument Specifications in meta/argument_specs</b></summary>
{% for section, specs in role.argument_specs.argument_specs.items() %}
#### Key: {{ section }}
**Description**: {{ specs.description or specs.short_description or 'No description provided' }}
{{ render_arguments_list(specs.options) }}
{% endfor %}
</details>
{% else %}
{% endif %}

{% macro render_repo_link(repo, role_name, file_path, line, repo_type, branch) -%}
  {%- if repo and file_path and line is not none -%}
    {%- if role.belongs_to_collection -%}
      {%- set full_path = 'roles/' ~ role_name ~ '/' ~ file_path -%}
    {%- else -%}
      {%- set full_path = file_path -%}
    {%- endif %}
    {%- set encoded_path = full_path | replace(' ', '%20') -%}

    {%- if repo_type == 'github' -%}
      {{ repo }}/blob/{{ branch }}/{{ encoded_path }}#L{{ line }}
    {%- elif repo_type == 'gitlab' -%}
      {{ repo }}/-/blob/{{ branch }}/{{ encoded_path }}#L{{ line }}
    {%- elif repo_type == 'gitea' -%}
      {{ repo }}/src/branch/{{ branch }}/{{ encoded_path }}#L{{ line }}
    {%- else -%}
      {{ repo }}/{{ encoded_path }}#L{{ line }}
    {%- endif %}
  {%- else -%}
    {{ file_path }}#L{{ line }}
  {%- endif %}
{%- endmacro %}

{% if role.defaults|length > 0 -%}
### Defaults

**These are static variables with lower priority**
{%- for defaultfile in role.defaults|sort(attribute='file') %}

#### File: defaults/{{ defaultfile.file }}
{# Cycle used for deciding to set Title and Required Column #}
{% set ns = namespace(details_required = false, details_title = false, details_choices = false) %}
{%- for key, details in defaultfile.data.items() -%}
    {%- if details.required is not none -%}{%- set ns.details_required = true -%}{%- endif -%}
    {%- if details.title is not none -%}{%- set ns.details_title = true -%}{%- endif -%}
    {%- if details.choices != None -%}{%- set ns.details_choices = true -%}{%- endif -%}
{%- endfor -%}
| Var          | Type         | Value       |{% if ns.details_choices %}Choices    |{% endif %}{% if ns.details_required %}Required    |{% endif %}{% if ns.details_title %} Title       |{% endif %}
|--------------|--------------|-------------|{% if ns.details_choices %}-----------|{% endif %}{% if ns.details_required %}------------|{% endif %}{% if ns.details_title %}-------------|{% endif %}
{%- for key, details in defaultfile.data.items() %}
{%- set var_type = details.value.__class__.__name__ %}
{%- if '.' in key %}
| [{{ key.rsplit('.', 1)[0] ~ '.**' ~ key.rsplit('.', 1)[1] ~ '**' }}]({{ render_repo_link(role.repository, role.name, 'defaults/' ~ defaultfile.file, details.line, role.repository_type, role.repository_branch) }})   | {{ var_type }} | {% if details.value is string and details.value | length == 0 %}{% else %}`{{ details.value | replace('|', '¦') }}`{% endif %} | {% if ns.details_choices %} {{ details.choices | replace('|', '¦') }}  |{% endif %}  {% if ns.details_required %} {{ details.required }}  |{% endif %} {% if ns.details_title %} {{ details.title | replace('|', '¦') }} |{% endif %}
{%- else %}
| [{{ key }}]({{ render_repo_link(role.repository, role.name, 'defaults/' ~ defaultfile.file, details.line, role.repository_type, role.repository_branch) }})   | {{ var_type }} | {% if details.value is string and details.value | length == 0 %}{% else %}`{{ details.value | replace('|', '¦') }}`{% endif %} | {% if ns.details_choices %} {{ details.choices | replace('|', '¦') }}  |{% endif %}  {% if ns.details_required %} {{ details.required }}  |{% endif %} {% if ns.details_title %} {{ details.title | replace('|', '¦') }} |{% endif %}
{%- endif %}
{%- endfor %}
{%- endfor %}

{%- for defaultfile in role.defaults|sort(attribute='file') -%}
{%- set ns = namespace(has_descriptions = false) -%}
{%- for key, details in defaultfile.data.items() -%}
    {%- if details.description != None -%}{%- set ns.has_descriptions = true -%}{% endif -%}
{%- endfor -%}
{%- if ns.has_descriptions %}
<details>
<summary><b>🖇️ Full descriptions for vars in defaults/{{ defaultfile.file }}</b></summary>
<br>
<table>
<th>Var</th><th>Description</th>
{%- for key, details in defaultfile.data.items() %}
    {%- if details.description != None %}
<tr><td><b>{{ key }}</b></td><td>{{details.description }}</td></tr>
    {%- endif %}
{%- endfor %}
</table>
<br>
</details>
{%- endif %}
{%- endfor %}
{%- else %}
{%- endif %}


{% if role.vars|length > 0 -%}
### Vars

**These are variables with higher priority**
{%- for varsfile in role.vars|sort(attribute='file') %}
#### File: vars/{{ varsfile.file }}
{# Cycle used for deciding to set Title and Required Column #}
{% set ns = namespace(details_required = false, details_title = false, details_choices = false) %}
{%- for key, details in varsfile.data.items() -%}
    {%- if details.required is not none -%}{%- set ns.details_required = true -%}{%- endif -%}
    {%- if details.title is not none -%}{%- set ns.details_title = true -%}{%- endif -%}
    {%- if details.choices != None -%}{%- set ns.details_choices = true -%}{%- endif -%}
{%- endfor -%}
| Var          | Type         | Value       |{% if ns.details_choices %}Choices    |{% endif %}{% if ns.details_required %}Required    |{% endif %}{% if ns.details_title %} Title       |{% endif %}
|--------------|--------------|-------------|{% if ns.details_choices %}-----------|{% endif %}{% if ns.details_required %}------------|{% endif %}{% if ns.details_title %}-------------|{% endif %}
{%- for key, details in varsfile.data.items() %}
{%- set var_type = details.value.__class__.__name__ %}
{%- if '.' in key %}
| [{{ key.rsplit('.', 1)[0] ~ '.**' ~ key.rsplit('.', 1)[1] ~ '**' }}]({{ render_repo_link(role.repository, role.name, 'vars/' ~ varsfile.file, details.line, role.repository_type, role.repository_branch) }})   | {{ var_type }} | {% if details.value is string and details.value | length == 0 %}{% else %}`{{ details.value | replace('|', '¦') }}`{% endif %} | {% if ns.details_choices %} {{ details.choices | replace('|', '¦') }}  |{% endif %}  {% if ns.details_required %} {{ details.required }}  |{% endif %} {% if ns.details_title %} {{ details.title | replace('|', '¦') }} |{% endif %}
{%- else %}
| [{{ key }}]({{ render_repo_link(role.repository, role.name, 'vars/' ~ varsfile.file, details.line, role.repository_type, role.repository_branch) }})   | {{ var_type }} | {% if details.value is string and details.value | length == 0 %}{% else %}`{{ details.value | replace('|', '¦') }}`{% endif %} | {% if ns.details_choices %} {{ details.choices | replace('|', '¦') }}  |{% endif %}  {% if ns.details_required %} {{ details.required }}  |{% endif %} {% if ns.details_title %} {{ details.title | replace('|', '¦') }} |{% endif %}
{%- endif %}
{%- endfor %}
{%- endfor %}

{%- for varsfile in role.vars|sort(attribute='file') -%}
{% set ns = namespace(has_descriptions = false) -%}
{%- for key, details in varsfile.data.items() -%}
    {%- if details.description != None -%}{%- set ns.has_descriptions = true -%}{%- endif %}
{%- endfor %}
{%- if ns.has_descriptions %}
<details>
<summary><b>🖇️ Full Descriptions for vars in vars/{{ varsfile.file }}</b></summary>
<br>
<table>
<th>Var</th><th>Description</th>
{%- for key, details in varsfile.data.items() %}
    {%- if details.description != None %}
<tr><td><b>{{ key }}</b></td><td>{{details.description }}</td></tr>
    {%- endif %}
{%- endfor %}
</table>
<br>
</details>
{%- endif %}
{%- endfor %}
{%- else %}
{%- endif %}


### Tasks

{% for taskfile in role.tasks|sort(attribute='file') %}
#### File: tasks/{{ taskfile.file }}
{% set ns = namespace (tags_required = false, comments_required = false) %}
{%- for t in taskfile['mermaid'] %}
  {%- if 'tags' in t and t['tags'] != "" %}{% set ns.tags_required = true %}{% endif %}
{%- endfor %}
{%- for comment in taskfile['comments'] %}
  {%- if comment != "" %}{% set ns.comments_required = true %}{% endif %}
{%- endfor %}
| Name | Module | Has Conditions |{% if ns.tags_required %} Tags |{% endif %}{% if ns.comments_required %} Comments |{% endif %}
| ---- | ------ | -------------- |{% if ns.tags_required %} -----|{% endif %}{% if ns.comments_required %} -------- |{% endif %}
{%- for task in taskfile.tasks %}
{%- if taskfile['lines'] | length > 0 %}
| [{{ task.name.replace("|", "¦") }}]({{ render_repo_link(role.repository, role.name, 'tasks/' ~ taskfile.file, taskfile['lines'][task.name], role.repository_type, role.repository_branch) }}) | {{ task.module }} | {{ 'True' if task.when else 'False' }} |{% if ns.tags_required %} {{ taskfile['mermaid'] | selectattr('name', 'equalto', task.name) | map(attribute='tags') | list | first | join(',') }} |{% endif %}{% if ns.comments_required %} {{ taskfile['comments'] | selectattr('task_name', 'equalto', task.name) | map(attribute='task_comments') | join }} |{% endif %}
{%- else %}
| {{ task.name.replace("|", "¦") }} | {{ task.module }} | {{ 'True' if task.when else 'False' }} |{% if ns.tags_required %} {{ taskfile['mermaid'] | selectattr('name', 'equalto', task.name) | map(attribute='tags') | list | first | join(',') }} |{% endif %}{% if ns.comments_required %} {{ taskfile['comments'] | selectattr('task_name', 'equalto', task.name) | map(attribute='task_comments') | join }} |{% endif %}
{%- endif %}
{%- endfor %}
{% endfor %}

{% if mermaid_code_per_file -%}
## Task Flow Graphs

{% for task_file, mermaid_code in mermaid_code_per_file.items() %}

### Graph for {{ task_file }}

```mermaid
{{ mermaid_code }}
```
{% endfor %}
{%- endif %}

{% if role.playbook.content -%}
## Playbook

```yml
{{ role.playbook.content }}
```
{%- endif %}
{% if role.playbook.graph -%}
## Playbook graph
```mermaid
{{ role.playbook.graph }}
```
{%- endif %}

{% if role.meta.galaxy_info -%}
## Author Information
{{ role.meta.galaxy_info.author or 'Unknown Author' }}

#### License

{{ role.meta.galaxy_info.license or 'No license specified.' }}

#### Minimum Ansible Version

{{ role.meta.galaxy_info.min_ansible_version or 'No minimum version specified.' }}

#### Platforms

{% if role.meta.galaxy_info.platforms -%}
{% for platform in role.meta.galaxy_info.platforms -%}
- **{{ platform.name }}**: {{ platform.versions }}
{% endfor -%}
{%- else -%}
No platforms specified.
{%- endif %}
{%- endif %}

#### Dependencies

{% if role.meta.dependencies %}
{%- for dep in role.meta.dependencies -%}
- {% if dep is string -%}
  **{{ dep }}**
{%- elif dep.role -%}
  **{{ dep.role }}**{% if dep.name %} (alias: _{{ dep.name }}_){% endif -%}{% if dep.version %} - version: `{{ dep.version }}`{% endif -%}
  {% if dep.vars -%}
    {%- set vars = dep.vars.items() | list -%}
    {% if vars|length > 0 %}
    - **Vars:**
      {% for k, v in vars -%}
      - {{ k }} = {{ v }}
      {% endfor -%}
    {%- endif -%}
  {% endif %}
  {% if dep.tags -%}
    {% set tags = dep.tags if dep.tags is iterable and dep.tags is not string else [dep.tags] -%}
    - **Tags:** {{ tags | join(', ') }}
  {%- endif %}
  {% if dep.when -%}
    - **When:** `{{ dep.when }}`
  {%- endif %}
{%- else -%}
  {{ dep | tojson(indent=0) }}
{%- endif %}
{% endfor -%}
{%- else -%}
No dependencies specified.
{%- endif -%}
"""

collection_template = """
# 📃 Collection overview

**Namespace**: {{ collection.namespace }}

**Name**: {{ collection.name }}

**Version**: {{ collection.version }}

**Authors**:
{% for author in collection.authors %}
- {{ author }}\n
{%- endfor %}

{% if collection.description -%}
## Description

{{ collection.description }}
{%- endif %}

{% macro render_repo_role_readme_link(repo, role_name, repo_type, branch) -%}
  {%- if repo and role_name -%}
    {%- set file_path = 'roles/' ~ role_name -%}
    {%- set encoded_path = file_path | replace(' ', '%20') -%}
    {%- if repo_type == 'github' -%}
      {{ repo }}/tree/{{ branch }}/{{ encoded_path }}
    {%- elif repo_type == 'gitlab' -%}
      {{ repo }}/-/tree/{{ branch }}/{{ encoded_path }}
    {%- elif repo_type == 'gitea' -%}
      {{ repo }}/src/branch/{{ branch }}/{{ encoded_path }}
    {%- else -%}
      {{ repo }}/{{ encoded_path }}
    {%- endif %}
  {%- else -%}
    roles/{{ role_name }}
  {%- endif %}
{%- endmacro %}

## Roles
{% for role in roles|sort(attribute='name') %}
### [{{ role.name }}]({{ render_repo_role_readme_link(collection.repository, role.name, collection.repository_type, collection.repository_branch) }})
{% endfor %}
## Roles vars
{% for role in roles|sort(attribute='name') %}
# [{{ role.name }}]({{ render_repo_role_readme_link(collection.repository, role.name, collection.repository_type, collection.repository_branch) }})
{% if role.meta and role.meta.galaxy_info -%}
## {{ role.name }} Description:
{{ role.meta.galaxy_info.description or 'Not available.' }}
{% else %}
Description: Not available.
{%- endif %}
{% macro render_repo_link(repo, role_name, file_path, line, repo_type, branch) -%}
  {%- if repo and file_path and line is not none -%}
    {%- if role.belongs_to_collection -%}
      {%- set full_path = 'roles/' ~ role_name ~ '/' ~ file_path -%}
    {%- else -%}
      {%- set full_path = file_path -%}
    {%- endif %}
    {%- set encoded_path = full_path | replace(' ', '%20') -%}
    {%- if repo_type == 'github' -%}
      {{ repo }}/blob/{{ branch }}/{{ encoded_path }}#L{{ line }}
    {%- elif repo_type == 'gitlab' -%}
      {{ repo }}/-/blob/{{ branch }}/{{ encoded_path }}#L{{ line }}
    {%- elif repo_type == 'gitea' -%}
      {{ repo }}/src/branch/{{ branch }}/{{ encoded_path }}#L{{ line }}
    {%- else -%}
      {{ repo }}/{{ encoded_path }}#L{{ line }}
    {%- endif %}
  {%- else -%}
    {{ file_path }}#L{{ line }}
  {%- endif %}
{%- endmacro %}
{% macro render_arguments_list(arguments, level=0) %}
{% for arg, details in arguments.items() %}
  {%- set indent = '  ' * level %}
  {{ indent }}- **{{ arg }}**
  {{ indent }}  - **Required**: {{ details.required | default('false') }}
  {{ indent }}  - **Type**: {{ details.type }}
  {{ indent }}  - **Default**: {{ details.default | default('none') }}
  {{ indent }}  - **Description**: {{ details.description | default('No description provided') }}
  {% if details.choices is defined %}
    {{ indent }}  - **Choices**:
    {% for choice in details.choices %}
      {{ indent }}    - {{ choice }}
    {% endfor %}
  {% endif %}
  {% if details.aliases is defined %}
    {{ indent }}  - **Aliases**:
    {% for alias in details.aliases %}
      {{ indent }}    - {{ alias }}
    {% endfor %}
  {% endif %}
  {% if details.type == 'dict' and details.options %}
    {{ render_arguments_list(details.options, level + 1) }}
  {% elif details.type == 'list' and details.elements == 'dict' %}
    {% for elem in details.default %}
      {% if elem is mapping %}
        {{ render_arguments_list(elem, level + 1) }}
      {% endif %}
    {% endfor %}
  {% endif %}
{% endfor %}
{% endmacro %}
{% if role.argument_specs %}
<details>
<summary><b>🧩 {{ role.name }} Argument Specifications in meta/argument_specs</b></summary>
{% for section, specs in role.argument_specs.argument_specs.items() %}
#### Key: {{ section }}
**Description**: {{ specs.description or specs.short_description or 'No description provided' }}
{{ render_arguments_list(specs.options) }}
{% endfor %}
</details>
{% else %}
{% endif %}

{% if role.defaults|length > 0 -%}
### {{ role.name }} Defaults

**These are static variables with lower priority**
{%- for defaultfile in role.defaults|sort(attribute='file') %}

#### {{ role.name }} File: [defaults/{{ defaultfile.file }}]({{ render_repo_role_readme_link(collection.repository, role.name, collection.repository_type, collection.repository_branch) }}/defaults/{{ defaultfile.file }})
{# Cycle used for deciding to set Title and Required Column #}
{% set ns = namespace(details_required = false, details_title = false, details_choices = false) %}
{%- for key, details in defaultfile.data.items() -%}
    {%- if details.required is not none -%}{%- set ns.details_required = true -%}{%- endif -%}
    {%- if details.title is not none -%}{%- set ns.details_title = true -%}{%- endif -%}
    {%- if details.choices != None -%}{%- set ns.details_choices = true -%}{%- endif -%}
{%- endfor -%}
| Var          | Type         | Value       |{% if ns.details_choices %}Choices    |{% endif %}{% if ns.details_required %}Required    |{% endif %}{% if ns.details_title %} Title       |{% endif %}
|--------------|--------------|-------------|{% if ns.details_choices %}-------------|{% endif %}{% if ns.details_required %}-------------|{% endif %}{% if ns.details_title %}-------------|{% endif %}
{%- for key, details in defaultfile.data.items() %}
{%- set var_type = details.value.__class__.__name__ %}
| [{{ key }}]({{ render_repo_link(role.repository, role.name, 'defaults/' ~ defaultfile.file, details.line, role.repository_type, role.repository_branch) }})   | {{ var_type }}   | {% if details.value is string and details.value | length == 0 %}{% else %}`{{ details.value | replace('|', '¦') }}`{% endif %} | {% if ns.details_choices %} {{ details.choices | replace('|', '¦') }}  |{% endif %}  {% if ns.details_required %} {{ details.required }}  |{% endif %} {% if ns.details_title %} {{ details.title | replace('|', '¦') }} |{% endif %}
{%- endfor %}
{%- endfor %}

{%- for defaultfile in role.defaults|sort(attribute='file') -%}
{%- set ns = namespace(has_descriptions = false) -%}
{%- for key, details in defaultfile.data.items() -%}
    {%- if details.description != None -%}{%- set ns.has_descriptions = true -%}{% endif -%}
{%- endfor -%}
{%- if ns.has_descriptions %}
<details>
<summary><b>🖇️ {{ role.name }} Full descriptions for vars in defaults/{{ defaultfile.file }}</b></summary>
<br>
{%- for key, details in defaultfile.data.items() %}
    {%- if details.description != None %}
<b>{{ key }}:</b> {{ details.description }}
<br>
    {%- endif %}
{%- endfor %}
<br>
</details>
{%- endif %}
{%- endfor %}
{%- else %}
{%- endif %}


{% if role.vars|length > 0 -%}
### {{ role.name }} Vars

**These are variables with higher priority**
{%- for varsfile in role.vars|sort(attribute='file') %}
#### {{ role.name }} File: [vars/{{ varsfile.file }}]({{ render_repo_role_readme_link(collection.repository, role.name, collection.repository_type, collection.repository_branch) }}/vars/{{ varsfile.file }})
{# Cycle used for deciding to set Title and Required Column #}
{% set ns = namespace(details_required = false, details_title = false, details_choices = false) %}
{%- for key, details in varsfile.data.items() -%}
    {%- if details.required is not none -%}{%- set ns.details_required = true -%}{%- endif -%}
    {%- if details.title is not none -%}{%- set ns.details_title = true -%}{%- endif -%}
    {%- if details.choices != None -%}{%- set ns.details_choices = true -%}{%- endif -%}
{%- endfor -%}
| Var          | Type         | Value       |{% if ns.details_choices %}Choices    |{% endif %}{% if ns.details_required %}Required    |{% endif %}{% if ns.details_title %} Title       |{% endif %}
|--------------|--------------|-------------|{% if ns.details_choices %}-------------|{% endif %}{% if ns.details_required %}-------------|{% endif %}{% if ns.details_title %}-------------|{% endif %}
{%- for key, details in varsfile.data.items() %}
{%- set var_type = details.value.__class__.__name__ %}
| [{{ key }}]({{ render_repo_link(role.repository, role.name, 'vars/' ~ varsfile.file, details.line, role.repository_type, role.repository_branch) }})   | {{ var_type }}   | {% if details.value is string and details.value | length == 0 %}{% else %}`{{ details.value | replace('|', '¦') }}`{% endif %} | {% if ns.details_choices %} {{ details.choices | replace('|', '¦') }}  |{% endif %}  {% if ns.details_required %} {{ details.required }}  |{% endif %} {% if ns.details_title %} {{ details.title | replace('|', '¦') }} |{% endif %}
{%- endfor %}
{%- endfor %}

{%- for varsfile in role.vars|sort(attribute='file') -%}
{% set ns = namespace(has_descriptions = false) -%}
{%- for key, details in varsfile.data.items() -%}
    {%- if details.description != None -%}{%- set ns.has_descriptions = true -%}{%- endif %}
{%- endfor %}
{%- if ns.has_descriptions %}
<details>
<summary><b>🖇️ {{ role.name }} Full Descriptions for vars in vars/{{ varsfile.file }}</b></summary>
<br>
{%- for key, details in varsfile.data.items() %}
    {%- if details.description != None %}
<b>{{ key }}:</b> {{ details.description }}
<br>
    {%- endif %}
{%- endfor %}
<br>
</details>
{%- endif %}
{%- endfor %}
{%- else %}
{%- endif %}
{% endfor %}
## Metadata

{% if collection.repository -%}
- **Repository**: [Repository]({{ collection.repository }})
{% endif %}
{% if collection.documentation -%}
- **Documentation**: [Documentation]({{ collection.documentation }})
{% endif %}
{% if collection.homepage -%}
- **Homepage**: [Homepage]({{ collection.homepage }})
{% endif %}
{% if collection.issues -%}
- **Issues**: [Issues]({{ collection.issues }})
{% endif %}
"""



import re


def sanitize_for_mermaid_id(text):
    text = text.replace("|", "_")
    # Allowing a-zA-Z0-9 as well as French accents
    return re.sub(r'[^a-zA-Z0-9À-ÿ]', '_', text)


def break_text(text, max_length=50):
    words = text.split(' ')
    lines = []
    current_line = []
    current_length = 0
    for word in words:
        if current_length + len(word) + len(current_line) > max_length:
            lines.append(' '.join(current_line))
            current_length = 0
            current_line = []
        current_line.append(word)
        current_length += len(word)
    if current_line:
        lines.append(' '.join(current_line))
    return '<br>'.join(lines)


def sanitize_for_title(text):
    # Allowing a-z0-9 as well as French accents, and converting to lower case
    try:
        sanitized_text = re.sub(r'[^a-z0-9À-ÿ]', ' ', text.lower())
        return break_text(sanitized_text)
    except Exception as e:
        return "cannot handle"


def sanitize_for_condition(text, max_length=50):
    sanitized_text = re.sub(r'[^a-z0-9À-ÿ]', ' ', text.lower())
    return break_text(sanitized_text, max_length)


def process_tasks(tasks, last_node, mermaid_data, parent_node=None, level=0, in_rescue_block=False):
    for i, task in enumerate(tasks):
        has_rescue = False
        task_name = task.get("name", f"Unnamed_task_{i}")
        task_module_include_tasks = task.get(
            "ansible.builtin.include_tasks") or task.get("include_tasks", False)
        task_module_import_tasks = task.get(
            "ansible.builtin.import_tasks") or task.get("import_tasks", False)
        task_module_import_playbook = task.get(
            "ansible.builtin.import_playbook") or task.get("import_playbook", False)
        task_module_include_role = task.get(
            "ansible.builtin.include_role") or task.get("include_role", False)
        task_module_import_role = task.get(
            "ansible.builtin.import_role") or task.get("import_role", False)
        task_module_include_vars = task.get(
            "ansible.builtin.include_vars") or task.get("include_vars", False)
        when_condition = task.get("when", False)
        block = task.get("block", False)
        rescue = task.get("rescue", False)
        task_name = re.sub(r"{{\s*(\w+)\s*}}", r"\1", task_name)
        sanitized_task_name = sanitize_for_mermaid_id(f"{task_name}{i}")
        sanitized_task_title = sanitize_for_title(task_name)
        if when_condition:
            if isinstance(when_condition, list):
                when_condition = " AND ".join(when_condition)
            sanitized_when_condition = f"**{sanitize_for_condition(str(when_condition)).strip()}**"
            if 'When:' not in sanitized_task_title:
                sanitized_task_title += f'<br>When: {sanitized_when_condition}'
        if block:
            block_start_node = sanitized_task_name + f'_block_start_{level}'
            mermaid_data += f'\n  {last_node}-->|Block Start| {block_start_node}[[{sanitized_task_title}]]:::block'
            last_node, mermaid_data = process_tasks(
                block, block_start_node, mermaid_data, block_start_node, level + 1, in_rescue_block=False)
            if rescue:
                has_rescue = True
                rescue_start_node = sanitized_task_name + \
                    f'_rescue_start_{level}'
                mermaid_data += f'\n  {last_node}-->|Rescue Start| {rescue_start_node}[{sanitized_task_title}]:::rescue'
                last_node, mermaid_data = process_tasks(
                    rescue, rescue_start_node, mermaid_data, block_start_node, level + 1, in_rescue_block=True)
                end_label = "End of Rescue Block"
                mermaid_data += f'\n  {last_node}-.->|{end_label}| {block_start_node}'
        elif rescue:
            rescue_start_node = sanitized_task_name + f'_rescue_start_{level}'
            mermaid_data += f'\n  {last_node}-->|Rescue Start| {rescue_start_node}[{sanitized_task_title}]:::rescue'
            last_node, mermaid_data = process_tasks(
                rescue, rescue_start_node, mermaid_data, parent_node, level + 1, in_rescue_block=True)
            end_label = "End of Rescue Block"
            mermaid_data += f'\n  {last_node}-.->|{end_label}| {parent_node}'
        else:

            if task_module_include_tasks:
                if isinstance(task_module_include_tasks, dict):
                    check_style_included_tasks = task_module_include_tasks.get(
                        'file', task_module_include_tasks)
                else:
                    check_style_included_tasks = task_module_include_tasks
                sanitized_include_tasks_name = sanitize_for_mermaid_id(
                    f"{task_name}_{check_style_included_tasks}_{i}")
                sanitized_include_tasks_title = sanitize_for_title(
                    f"{check_style_included_tasks}")
                mermaid_data += f'\n  {last_node}-->|Include task| {sanitized_include_tasks_name}[{sanitized_task_title}<br>include_task: {sanitized_include_tasks_title}]:::includeTasks'
                last_node = sanitized_include_tasks_name

            elif task_module_import_tasks:
                if isinstance(task_module_import_tasks, dict):
                    check_style_imported_tasks = task_module_import_tasks.get(
                        'file', task_module_import_tasks)
                else:
                    check_style_imported_tasks = task_module_import_tasks
                sanitized_imported_tasks_name = sanitize_for_mermaid_id(
                    f"{task_name}_{check_style_imported_tasks}_{i}")
                sanitized_imported_tasks_title = sanitize_for_title(
                    f"{check_style_imported_tasks}")
                mermaid_data += f'\n  {last_node}-->|Import task| {sanitized_imported_tasks_name}[/{sanitized_task_title}<br>import_task: {sanitized_imported_tasks_title}/]:::importTasks'
                last_node = sanitized_imported_tasks_name

            elif task_module_import_playbook:
                if isinstance(task_module_import_playbook, dict):
                    check_style_import_playbook = task_module_import_playbook.get(
                        'file', task_module_import_playbook)
                else:
                    check_style_import_playbook = task_module_import_playbook
                sanitized_import_playbook_name = sanitize_for_mermaid_id(
                    f"{task_name}_{check_style_import_playbook}_{i}")
                sanitized_import_playbook_title = sanitize_for_title(
                    f"{check_style_import_playbook}")
                mermaid_data += f'\n  {last_node}-->|Import playbook| {sanitized_import_playbook_name}[/{sanitized_task_title}<br>import_playbook: {sanitized_import_playbook_title}/]:::importPlaybook'
                last_node = sanitized_import_playbook_name

            elif task_module_include_role:
                if isinstance(task_module_include_role, dict):
                    check_style_include_role = task_module_include_role.get(
                        'name', task_module_include_role)
                else:
                    check_style_include_role = task_module_include_role
                sanitized_include_role_name = sanitize_for_mermaid_id(
                    f"{task_name}_{check_style_include_role}_{i}")
                sanitized_include_role_title = sanitize_for_title(
                    check_style_include_role)
                mermaid_data += f'\n  {last_node}-->|Include role| {sanitized_include_role_name}({sanitized_task_title}<br>include_role: {sanitized_include_role_title}):::includeRole'
                last_node = sanitized_include_role_name

            elif task_module_import_role:
                if isinstance(task_module_import_role, dict):
                    check_style_import_role = task_module_import_role.get(
                        'name', task_module_import_role)
                else:
                    check_style_import_role = task_module_import_role
                sanitized_import_role_name = sanitize_for_mermaid_id(
                    f"{task_name}_{check_style_import_role}_{i}")
                sanitized_import_role_title = sanitize_for_title(
                    check_style_import_role)
                mermaid_data += f'\n  {last_node}-->|Import role| {sanitized_import_role_name}([{sanitized_task_title}<br>import_role: {sanitized_import_role_title}]):::importRole'
                last_node = sanitized_import_role_name

            elif task_module_include_vars:
                if isinstance(task_module_include_vars, dict):
                    check_style_include_vars = task_module_include_vars.get(
                        'file', False) or task_module_include_vars.get('dir', task_module_include_vars)
                else:
                    check_style_include_vars = task_module_include_vars
                sanitized_include_vars_name = sanitize_for_mermaid_id(
                    f"{task_name}_{check_style_include_vars}_{i}")
                sanitized_include_vars_title = sanitize_for_title(
                    check_style_include_vars)
                mermaid_data += f'\n  {last_node}-->|Include vars| {sanitized_include_vars_name}[{sanitized_task_title}<br>include_vars: {sanitized_include_vars_title}]:::includeVars'
                last_node = sanitized_include_vars_name

            else:
                mermaid_data += f'\n  {last_node}-->|Task| {sanitized_task_name}[{sanitized_task_title}]:::task'
                last_node = sanitized_task_name

    if parent_node and not in_rescue_block and not has_rescue:
        end_label = "End of Block"
        mermaid_data += f'\n  {last_node}-.->|{end_label}| {parent_node}'

    return last_node, mermaid_data


def generate_mermaid_playbook(playbook):
    mermaid_data = "flowchart TD"
    for play in playbook:
        hosts = play.get("hosts", "UndefinedHost")
        tasks = play.get("tasks", [])
        roles = play.get("roles", [])
        if not isinstance(hosts, list):
            hosts = [hosts]
        sanitized_hosts = []
        for host in hosts:
            host = re.sub(r"{{\s*(\w+)\s*}}", r"\1", host)
            host = sanitize_for_mermaid_id(host)
            sanitized_hosts.append(host)
        sanitized_hosts = ", ".join(sanitized_hosts)
        sanitized_hosts = "hosts[" + sanitized_hosts + "]"
        last_node = sanitized_hosts
        if roles:
            for i, role in enumerate(roles):
                role_name = role["role"] if isinstance(role, dict) else role
                role_name = role_name if role_name else f"Unnamed_role_{i}"
                role_name = re.sub(r"{{\s*(\w+)\s*}}", r"\1", role_name)
                sanitized_role_name = sanitize_for_mermaid_id(role_name)
                sanitized_role_title = sanitize_for_title(role_name)
                mermaid_data += f'\n  {last_node}-->|Role| {sanitized_role_name}[{sanitized_role_title}]'
                last_node = sanitized_role_name
        last_node, mermaid_data = process_tasks(tasks, last_node, mermaid_data)
    return mermaid_data


def generate_mermaid_role_tasks_per_file(tasks_per_file):
    mermaid_codes = {}
    for task_info in tasks_per_file:
        task_file = task_info['file']
        tasks = task_info['mermaid']
        mermaid_data = """flowchart TD
Start
classDef block stroke:#3498db,stroke-width:2px;
classDef task stroke:#4b76bb,stroke-width:2px;
classDef includeTasks stroke:#16a085,stroke-width:2px;
classDef importTasks stroke:#34495e,stroke-width:2px;
classDef includeRole stroke:#2980b9,stroke-width:2px;
classDef importRole stroke:#699ba7,stroke-width:2px;
classDef includeVars stroke:#8e44ad,stroke-width:2px;
classDef rescue stroke:#665352,stroke-width:2px;
"""
        last_node = "Start"
        last_node, mermaid_data = process_tasks(tasks, last_node, mermaid_data)
        mermaid_data += f'\n  {last_node}-->End'
        mermaid_codes[task_file] = mermaid_data

    return mermaid_codes



# Import libraries
import os
import yaml
import click
from shutil import copyfile
from datetime import datetime
from jinja2 import Environment, BaseLoader, FileSystemLoader
from docsible.markdown_template import static_template, collection_template
from docsible.utils.mermaid import generate_mermaid_playbook, generate_mermaid_role_tasks_per_file
from docsible.utils.yaml import load_yaml_generic, load_yaml_files_from_dir_custom, get_task_comments, get_task_line_numbers
from docsible.utils.special_tasks_keys import process_special_task_keys
from docsible.utils.git import get_repo_info

DOCSIBLE_START_TAG = "<!-- DOCSIBLE START -->"
DOCSIBLE_END_TAG = "<!-- DOCSIBLE END -->"
timestamp = datetime.now().strftime('%Y%m%d%H%M%S')


def get_version():
    return "0.7.25"


def manage_docsible_file_keys(docsible_path):
    default_data = {
        'description': None,
        'requester': None,
        'users': None,
        'dt_dev': None,
        'dt_prod': None,
        'dt_update': datetime.now().strftime('%d/%m/%Y'),
        'version': None,
        'time_saving': None,
        'category': None,
        'subCategory': None,
        'aap_hub': None,
        'critical': None,
        'automation_kind': None
    }
    if os.path.exists(docsible_path):
        with open(docsible_path, 'r') as f:
            existing_data = yaml.safe_load(f) or {}

        updated_data = {**default_data, **existing_data}
        if updated_data != existing_data:
            # Update dt_update field if docsible keys added
            existing_data['dt_update'] = datetime.now().strftime('%d/%m/%Y')
            with open(docsible_path, 'w', encoding='utf-8') as f:
                yaml.dump(updated_data, f, default_flow_style=False)
            print(f"Updated {docsible_path} with new keys.")
    else:
        with open(docsible_path, 'w', encoding='utf-8') as f:
            yaml.dump(default_data, f, default_flow_style=False)
        print(f"Initialized {docsible_path} with default keys.")


def manage_docsible_tags(content):
    return f"{DOCSIBLE_START_TAG}\n{content}\n{DOCSIBLE_END_TAG}\n"


def replace_between_tags(existing_content, new_content):
    start_index = existing_content.find(DOCSIBLE_START_TAG)
    end_index = existing_content.find(DOCSIBLE_END_TAG) + len(DOCSIBLE_END_TAG)

    if start_index != -1 and end_index != -1:
        before = existing_content[:start_index].rstrip()
        after = existing_content[end_index:].lstrip()
        return f"{before}\n{new_content}\n{after}".strip()
    else:
        return existing_content + '\n' + new_content.strip()


def render_readme_template(collection_metadata, md_collection_template, roles_info, output_path, append):
    """
    Render the collection README.md using an embedded Jinja template.
    """
    # Render the static template
    if md_collection_template:
        template_dir = os.path.dirname(md_collection_template)
        template_file = os.path.basename(md_collection_template)
        env = Environment(loader=FileSystemLoader(template_dir))
        template = env.get_template(template_file)
    else:
        env = Environment(loader=BaseLoader)
        template = env.from_string(collection_template)
    data = {
        'collection': collection_metadata,
        'roles': roles_info
    }
    new_content = template.render(data)
    new_content = manage_docsible_tags(new_content)

    if os.path.exists(output_path):
        with open(output_path, 'r', encoding='utf-8') as f:
            existing_readme = f.read()
        if append:
            if DOCSIBLE_START_TAG in existing_readme and DOCSIBLE_END_TAG in existing_readme:
                final_content = replace_between_tags(
                    existing_readme, new_content)
            else:
                final_content = f"{existing_readme}\n{new_content}"
        else:
            final_content = new_content
    else:
        final_content = new_content

    with open(output_path, 'w', encoding='utf-8') as readme_file:
        readme_file.write(final_content)
    print(f"Collection README.md written at: {output_path}")


def document_collection_roles(collection_path, playbook, graph, no_backup, no_docsible, comments, task_line, md_collection_template, md_role_template, append, output, repository_url, repo_type, repo_branch):
    """
    Document all roles in a collection, extracting metadata from galaxy.yml or galaxy.yaml.
    """

    try:
        git_info = get_repo_info(collection_path) or {}
    except Exception as e:
        print(f"[WARN] Could not get Git info: {e}")
        git_info = {}
    repository_url = git_info.get("repository") if git_info else None
    repo_branch = repo_branch or (
        git_info.get("branch") if git_info else "main")
    repo_type = repo_type or (
        git_info.get("repository_type") if git_info else None)

    for root, dirs, files in os.walk(collection_path):
        galaxy_file = next(
            (f for f in files if f in ['galaxy.yml', 'galaxy.yaml']), None)
        if galaxy_file:
            galaxy_path = os.path.join(root, galaxy_file)
            with open(galaxy_path, 'r') as f:
                collection_metadata = yaml.safe_load(f)
                if output == "README.md":
                    readme_path = os.path.join(
                        root, collection_metadata.get('readme', output))
                else:
                    readme_path = os.path.join(root, output)

            collection_metadata['repository_type'] = repo_type
            collection_metadata['repository_branch'] = repo_branch

            if os.path.exists(readme_path) and not no_backup:
                backup_path = f"{readme_path[:readme_path.lower().rfind('.md')]}_backup_{timestamp}.md"
                copyfile(readme_path, backup_path)
                print(f"Backup of existing {output} created at: {backup_path}")

            roles_dir = os.path.join(root, 'roles')
            roles_info = []
            if os.path.exists(roles_dir) and os.path.isdir(roles_dir):
                for role in os.listdir(roles_dir):
                    role_path = os.path.join(roles_dir, role)
                    if os.path.isdir(role_path):
                        if playbook:
                            playbook_content = None
                            role_playbook_path = os.path.join(
                                role_path, playbook)
                            try:
                                with open(role_playbook_path, 'r') as f:
                                    playbook_content = f.read()
                            except FileNotFoundError:
                                print(f'{role} not found:', role_playbook_path)
                            except Exception as e:
                                print(f'{playbook} import for {role} error:', e)
                        role_info = document_role(role_path, playbook_content, graph, no_backup, no_docsible, comments, task_line, md_role_template,
                                                  belongs_to_collection=collection_metadata, append=append, output=output, repository_url=repository_url, repo_type=repo_type, repo_branch=repo_branch)
                        roles_info.append(role_info)

            render_readme_template(
                collection_metadata, md_collection_template, roles_info, readme_path, append)


@click.command()
@click.option('--role', '-r', default=None, help='Path to the Ansible role directory.')
@click.option('--collection', '-c', default=None, help='Path to the Ansible collection directory.')
@click.option('--playbook', '-p', default='tests/test.yml', help='Path to the playbook file.')
@click.option('--graph', '-g', is_flag=True, help='Generate Mermaid graph for tasks.')
@click.option('--no-backup', '-nob', is_flag=True, help='Do not backup the readme before remove.')
@click.option('--no-docsible', '-nod', is_flag=True, help='Do not generate .docsible file and do not include it in README.md.')
@click.option('--comments', '-com', is_flag=True, help='Read comments from tasks files')
@click.option('--task-line', '-tl', is_flag=True, help='Read line numbers from tasks')
@click.option('--md-collection-template', '-ctpl', default=None, help='Path to the collection markdown template file.')
@click.option('--md-role-template', '-rtpl', '--md-template', '-tpl', default=None, help='Path to the role markdown template file.')
@click.option('--append', '-a', is_flag=True, help='Append to the existing README.md instead of replacing it.')
@click.option('--output', '-o', default='README.md', help='Output readme file name.')
@click.option('--repository-url', '-ru', default=None, help='Repository base URL (used for standalone roles)')
@click.option('--repo-type', '-rt', default=None, help='Repository type: github, gitlab, gitea, etc.')
@click.option('--repo-branch', '-rb', default=None, help='Repository branch name (e.g., main or master)')
@click.version_option(version=get_version(), help=f"Show the module version. Actual is {get_version()}")
def doc_the_role(role, collection, playbook, graph, no_backup, no_docsible, comments, task_line, md_collection_template, md_role_template, append, output, repository_url, repo_type, repo_branch):
    if collection:
        collection_path = os.path.abspath(collection)
        if not os.path.exists(collection_path) or not os.path.isdir(collection_path):
            print(f"Folder {collection_path} does not exist.")
            return
        document_collection_roles(collection_path, playbook, graph, no_backup, no_docsible, comments, task_line,
                                  md_collection_template, md_role_template, append, output, repository_url, repo_type, repo_branch)
    elif role:
        role_path = os.path.abspath(role)
        if not os.path.exists(role_path) or not os.path.isdir(role_path):
            print(f"Folder {role_path} does not exist.")
            return

        if playbook == 'tests/test.yml':
            playbook = os.path.join(role_path, playbook)

        playbook_content = None
        if playbook:
            try:
                with open(playbook, 'r') as f:
                    playbook_content = f.read()
            except FileNotFoundError:
                print('playbook not found:', playbook)
            except Exception as e:
                print('playbook import error:', e)
        document_role(role_path, playbook_content, graph, no_backup, no_docsible, comments, task_line, md_role_template, belongs_to_collection=False,
                      append=append, output=output, repository_url=repository_url, repo_type=repo_type, repo_branch=repo_branch)
    else:
        print("Please specify either a role or a collection path.")


def document_role(role_path, playbook_content, generate_graph, no_backup, no_docsible, comments, task_line, md_role_template, belongs_to_collection, append, output, repository_url, repo_type, repo_branch):
    role_name = os.path.basename(role_path)
    readme_path = os.path.join(role_path, output)
    meta_path = os.path.join(role_path, "meta", "main.yml")
    docsible_path = os.path.join(role_path, ".docsible")
    argument_specs_path = os.path.join(role_path, "meta", "argument_specs.yml")

    if not no_docsible:
        manage_docsible_file_keys(docsible_path)

    # Check if meta/main.yml exist, otherwise try meta/main.yaml
    if not os.path.exists(meta_path):
        meta_path = os.path.join(role_path, "meta", "main.yaml")

    if not os.path.exists(argument_specs_path):
        argument_specs_path = os.path.join(
            role_path, "meta", "argument_specs.yaml")

    if os.path.exists(argument_specs_path):
        argument_specs = load_yaml_generic(argument_specs_path)
    else:
        argument_specs = None

    defaults_data = load_yaml_files_from_dir_custom(
        os.path.join(role_path, "defaults")) or []
    vars_data = load_yaml_files_from_dir_custom(
        os.path.join(role_path, "vars")) or []

    if repository_url == "detect":
        try:
            git_info = get_repo_info(role_path) or {}
        except Exception as e:
            print(f"[WARN] Could not get Git info: {e}")
            git_info = {}
        repository_url = git_info.get("repository") if git_info else None
        repo_branch = repo_branch or (
            git_info.get("branch") if git_info else "main")
        repo_type = repo_type or (
            git_info.get("repository_type") if git_info else None)

    role_info = {
        "name": role_name,
        "defaults": defaults_data,
        "vars": vars_data,
        "tasks": [],
        "meta": load_yaml_generic(meta_path) or {},
        "playbook": {"content": playbook_content, "graph":
                     generate_mermaid_playbook(yaml.safe_load(playbook_content)) if generate_graph and playbook_content else None},
        "docsible": load_yaml_generic(docsible_path) if not no_docsible else None,
        "belongs_to_collection": belongs_to_collection,
        "repository": repository_url,
        "repository_type": repo_type,
        "repository_branch": repo_branch,
        "argument_specs": argument_specs
    }

    tasks_dir = os.path.join(role_path, "tasks")
    role_info["tasks"] = []

    if os.path.exists(tasks_dir) and os.path.isdir(tasks_dir):
        for dirpath, dirnames, filenames in os.walk(tasks_dir):
            for task_file in filenames:
                if task_file.endswith(".yml") or task_file.endswith(".yaml"):
                    file_path = os.path.join(dirpath, task_file)
                    tasks_data = load_yaml_generic(file_path)
                    if tasks_data:
                        relative_path = os.path.relpath(file_path, tasks_dir)
                        task_info = {'file': relative_path,
                                     'tasks': [], 'mermaid': [], "comments": [], 'lines': []}
                        if comments:
                            task_info['comments'] = get_task_comments(
                                file_path)
                        if task_line:
                            task_info['lines'] = get_task_line_numbers(
                                file_path)
                        if not isinstance(tasks_data, list):
                            print(
                                f"Unexpected data type for tasks in {task_file}. Skipping.")
                            continue
                        for task in tasks_data:
                            if not isinstance(task, dict):
                                print(
                                    f"Skipping unexpected data in {task_file}: {task}")
                                continue
                            if task and len(task.keys()) > 0:
                                processed_tasks = process_special_task_keys(
                                    task)
                                task_info['tasks'].extend(processed_tasks)
                                task_info['mermaid'].extend([task])

                        role_info["tasks"].append(task_info)

    if os.path.exists(readme_path):
        if not no_backup:
            backup_readme_path = os.path.join(
                role_path, f"{output[:output.lower().rfind('.md')]}_backup_{timestamp}.md")
            copyfile(readme_path, backup_readme_path)
            print(f'Readme file backed up as: {backup_readme_path}')
        if not append:
            os.remove(readme_path)

    mermaid_code_per_file = {}
    if generate_graph:
        mermaid_code_per_file = generate_mermaid_role_tasks_per_file(
            role_info["tasks"])

    # Render the static template
    if md_role_template:
        template_dir = os.path.dirname(md_role_template)
        template_file = os.path.basename(md_role_template)
        env = Environment(loader=FileSystemLoader(template_dir))
        template = env.get_template(template_file)
    else:
        env = Environment(loader=BaseLoader)
        template = env.from_string(static_template)
    new_content = template.render(
        role=role_info, mermaid_code_per_file=mermaid_code_per_file)
    new_content = manage_docsible_tags(new_content)

    if os.path.exists(readme_path):
        with open(readme_path, 'r', encoding='utf-8') as f:
            existing_readme = f.read()
        if append:
            if DOCSIBLE_START_TAG in existing_readme and DOCSIBLE_END_TAG in existing_readme:
                final_content = replace_between_tags(
                    existing_readme, new_content)
            else:
                final_content = f"{existing_readme}\n{new_content}"
        else:
            final_content = new_content
    else:
        final_content = new_content

    with open(readme_path, "w", encoding='utf-8') as f:
        f.write(final_content)

    print('Documentation generated at:', readme_path)
    return role_info


if __name__ == '__main__':
    doc_the_role()



"""Module providing yaml parsing functions"""
import os
import yaml
import re


def vault_constructor(loader, node):
    # Handle '!vault' tag to prevent constructor issues
    return "ENCRYPTED_WITH_ANSIBLE_VAULT"


# Register the custom constructor with the '!vault' tag.
yaml.SafeLoader.add_constructor('!vault', vault_constructor)


def load_yaml_generic(filepath):
    """Function to load YAML in a standard way"""
    try:
        with open(filepath, 'r', encoding='utf-8') as f:
            data = yaml.safe_load(f)
        return data
    except (FileNotFoundError, yaml.constructor.ConstructorError) as e:
        print(f"Error loading {filepath}: {e}")
        return None


def get_multiline_indicator(line):
    """
    Detect and map YAML multiline scalar indicators to a descriptive name.
    Handles all combinations of |, >, +, -, and 1-9 indent levels.
    Returns: e.g., 'literal', 'folded_keep_indent_2', or 'invalid_...'
    """
    match = re.match(r'^\s*\w[\w\-\.]*\s*:\s*([>|][^\s#]*)', line)
    if not match:
        return None

    raw = match.group(1)

    # Invalid if multiple digits or unknown characters
    if re.search(r'\d{2,}', raw) or re.search(r'[^>|0-9+-]', raw):
        return f'invalid_{raw}'

    style = 'literal' if raw.startswith('|') else 'folded'
    chomping = None
    indent = None

    # Extract components
    chomp_match = re.search(r'[+-]', raw)
    indent_match = re.search(r'[1-9]', raw)

    if chomp_match:
        chomping = chomp_match.group(0)
    if indent_match:
        indent = indent_match.group(0)

    # Build result
    name = style
    if chomping == '+':
        name += '_keep'
    elif chomping == '-':
        name += '_strip'
    if indent:
        name += f'_indent_{indent}'

    return name


def load_yaml_file_custom(filepath):
    """
    Load a YAML file and extract both its data and associated metadata from comments, 
    while also tracking the line number for each key and nested item.
    The function parses the YAML file, collects metadata (title, required, choices, description) 
    from preceding comments for each key, and tracks the line number where each key appears. 
    It supports nested dictionaries and lists, and can handle multi-line values and 
    extended descriptions via special comment blocks.
    Args:
        filepath (str): Path to the YAML file.
    Returns:
        dict or None: A dictionary mapping each key path to its value, metadata, and line number,
        or None if the file is empty or an error occurs.
    """
    try:
        with open(filepath, 'r', encoding='utf-8') as file:
            lines = file.readlines()

        with open(filepath, 'r', encoding='utf-8') as file:
            data = yaml.safe_load(file)

        if not data:
            return None

        result = {}
        parent_line = 0

        def is_multiline_value(line):
            """
            Determine if a line in the YAML file indicates the start of a multi-line value.
            Args:
                line (str): A line from the YAML file.
            Returns:
                bool: True if the line ends with '|', indicating a multi-line value.
            """
            return line.strip().endswith('|')

        def extract_metadata(idx):
            """
            Extract metadata (title, required, choices, description) from comments preceding a given line index.
            Args:
                idx (int or None): The line index in the file for the current key.
            Returns:
                dict: Metadata dictionary with keys 'title', 'required', 'choices', and 'description'.
            """
            if idx is None:
                return {'title': None, 'required': None, 'choices': None, 'description': None}

            comments = []
            for comment_index in range(idx - 1, -1, -1):
                line = lines[comment_index].strip()
                if line.startswith("#"):
                    comments.append(line[1:].strip())
                else:
                    break
            comments.reverse()

            meta = {'title': None, 'required': None, 'choices': None, 'description': None}

            for comment in comments:
                lc = comment.lower()
                if lc.startswith('title:'):
                    meta['title'] = comment[6:].strip()
                elif lc.startswith('required:'):
                    meta['required'] = comment[9:].strip()
                elif lc.startswith('choices:'):
                    meta['choices'] = comment[8:].strip()
                elif lc.startswith('description:'):
                    meta['description'] = comment[12:].strip()
                elif lc.startswith('description-lines:'):
                    description_lines = []
                    start_collecting = False  # Flag to start collecting lines

                    # Process all subsequent lines to collect description content
                    for subsequent_line in lines[comment_index:]:
                        line_content = subsequent_line.strip()

                        # Start collecting after `description-lines:`
                        if line_content.startswith("#") and 'description-lines:' in line_content:
                            start_collecting = True
                            continue

                        # Stop collecting when encountering `# end`
                        if line_content.startswith("# end"):
                            break
                        
                        if start_collecting:
                            if line_content.startswith("#"):
                                # Collect the line content
                                description_lines.append(
                                    f'{line_content[1:].strip()}<br>')
                            else:
                                break  # Stop if a non-comment line is encountered
                    
                    # Join all collected lines into a single description string
                    if description_lines:
                        meta['description'] = "\n".join(
                            description_lines)
            return meta

        def process_line(k, v):
            """
            Process a single key-value pair, determine its line number, extract metadata, 
            and store the result in the output dictionary.
            Args:
                k (str): The full key path (dot-separated for nested keys).
                v (Any): The value associated with the key.
            """
            nonlocal parent_line

            line_idx = None
            dictkey = k.split('.')[-1]
            vtype = type(v)

            for idx in range(parent_line, len(lines)):
                line_stripped = lines[idx].strip()

                if line_stripped.startswith("#"):
                    continue

                if isinstance(v, dict):
                    dictvalue = None
                    if dictkey.isnumeric():
                        prev_path = ".".join(k.split(".")[:-1])
                        if result.get(prev_path, {}).get("type") == "list":
                            dictkey = list(v.keys())[0]
                            dictvalue = str(list(v.values())[0])
                    if dictvalue is None:
                        # dict
                        if line_stripped.startswith(f"{dictkey}:") or line_stripped.startswith(f"- {dictkey}:"):
                            line_idx = idx
                            break
                    else:
                        # inline dict
                        if f"{dictkey}:" in line_stripped and dictvalue in line_stripped:
                            line_idx = idx
                            break

                elif isinstance(v, list):
                    # list
                    if line_stripped.startswith(f"{dictkey}:") or line_stripped.startswith(f"- {v}"):
                        line_idx = idx
                        break

                else:
                    # key match
                    if line_stripped.startswith(f"{dictkey}:") or f"{dictkey}:" in line_stripped:
                        line_idx = idx
                        break
                    # bool in list
                    if vtype is bool and f"- {str(v).lower()}" in line_stripped.lower():
                        line_idx = idx
                        break
                    # none / null in list
                    if v is None and any(null_str in line_stripped.lower() for null_str in ['- none', '- null']):
                        line_idx = idx
                        break
                    # list item part 1
                    if f"- {str(v).lower()}" in line_stripped.lower():
                        line_idx = idx
                        break
                    # list item part 2
                    if dictkey.isnumeric():
                        prev_path = ".".join(k.split(".")[:-1])
                        if result.get(prev_path, {}).get("type") == "list":
                            if str(v).lower() in line_stripped.lower():
                                line_idx = idx
                                break

            current_line = line_idx if line_idx is not None else parent_line
            parent_line = current_line

            meta = extract_metadata(current_line)
            indicator_name = get_multiline_indicator(lines[current_line])
            result[k] = {
                'value': f"<multiline value: {indicator_name}>" if indicator_name
                        else [] if isinstance(v, list)
                        else {} if isinstance(v, dict)
                        else v.strip() if isinstance(v, str)
                        else v,
                'multiline_indicator': indicator_name,
                'title': meta['title'],
                'required': meta['required'],
                'choices': meta['choices'],
                'description': meta['description'],
                'line': current_line + 1,
                'type': 'dict' if isinstance(v, dict)
                        else 'list' if isinstance(v, list)
                        else type(v).__name__
            }

        def process_dict(base_key, value):
            """
            Recursively process a dictionary, handling each key-value pair and their nested structures.
            Args:
                base_key (str): The base key path for the current dictionary.
                value (dict): The dictionary to process.
            """
            for k, v in value.items():
                full_key = f"{base_key}.{k}"
                process_line(full_key, v)
                if isinstance(v, dict):
                    process_dict(full_key, v)
                elif isinstance(v, list):
                    process_list(full_key, v)

        def process_list(base_key, value):
            """
            Recursively process a list, handling each item and their nested structures.
            Args:
                base_key (str): The base key path for the current list.
                value (list): The list to process.
            """
            for idx, item in enumerate(value):
                full_key = f"{base_key}.{idx}"
                process_line(full_key, item)
                if isinstance(item, dict):
                    process_dict(full_key, item)
                elif isinstance(item, list):
                    process_list(full_key, item)

        for key, value in data.items():
            process_line(key, value)
            if isinstance(value, dict):
                process_dict(key, value)
            elif isinstance(value, list):
                process_list(key, value)

        return result

    except (FileNotFoundError, yaml.constructor.ConstructorError, yaml.YAMLError) as e:
        print(f"Error loading {filepath}: {e}")
        return None


def load_yaml_files_from_dir_custom(dir_path):
    """Function to load all YAML files from a given directory and include file names"""
    collected_data = []

    def process_yaml_file(full_path, dir_path):
        if full_path.endswith((".yml", ".yaml")):
            file_data = load_yaml_file_custom(full_path)
            if file_data:
                relative_path = os.path.relpath(full_path, dir_path)
                return ({'file': relative_path, 'data': file_data})
        return None

    if os.path.exists(dir_path) and os.path.isdir(dir_path):
        # dir-path
        for file in os.listdir(dir_path):
            full_path = os.path.join(dir_path, file)
            if os.path.isfile(full_path):
                item = process_yaml_file(full_path, dir_path)
                if item:
                    collected_data.append(item)
        # main-dir
        main_dir = os.path.join(dir_path, "main")
        if os.path.exists(main_dir) and os.path.isdir(main_dir):
            for root, _, files in os.walk(main_dir):
                for yaml_file in files:
                    full_path = os.path.join(root, yaml_file)
                    item = process_yaml_file(full_path, dir_path)
                    if item:
                        collected_data.append(item)

    return collected_data


def get_task_comments(filepath: str) -> list[dict[str, str]]:
    """
    Extracts comments for Ansible tasks.
    - For any named task (block or regular), uses any immediately preceding comments.
    - A blank line between a comment and the next task (or its direct comments)
      will prevent the earlier comment from being associated with that task.
    - Ignores comments not immediately preceding a named task due to other
      intervening non-comment lines.
    - Handles task names containing '#' if they are quoted, while still removing
      actual inline comments.
    """
    try:
        with open(filepath, 'r', encoding='utf-8') as f:
            lines = f.readlines()
    except FileNotFoundError:
        print(f"Error: File not found {filepath}")
        return []
    except Exception as e:
        print(f"Error reading file {filepath}: {e}")
        return []

    output_task_comments = []
    # This list will hold comment lines gathered immediately before a potential task.
    candidate_comments = []

    for i, line_content in enumerate(lines):
        stripped_line = line_content.strip()

        if stripped_line.startswith("#"):
            # Collect comments that might belong to the next task
            candidate_comments.append(stripped_line[1:].strip())
        elif stripped_line.startswith("- name:"):
            # This line defines a task. Process it and its collected candidate_comments.
            try:
                # Get the entire string part after "- name:"
                task_name_value_and_inline_comment = stripped_line.split(":", 1)[1]

                # Robustly find the first '#' that is NOT within quotes
                # to separate the actual task name value from a true inline comment.
                in_single_quote = False
                in_double_quote = False
                name_part_end_index = len(task_name_value_and_inline_comment)

                for k, char_code in enumerate(task_name_value_and_inline_comment):
                    # Basic quote state machine (doesn't handle escaped quotes within quotes)
                    if char_code == "'" and (k == 0 or task_name_value_and_inline_comment[k-1] != '\\'):
                        in_single_quote = not in_single_quote
                    elif char_code == '"' and (k == 0 or task_name_value_and_inline_comment[k-1] != '\\'):
                        in_double_quote = not in_double_quote
                    elif char_code == '#' and not in_single_quote and not in_double_quote:
                        name_part_end_index = k  # Found start of a true inline comment
                        break

                task_name_raw = task_name_value_and_inline_comment[:name_part_end_index].strip(
                )

            except IndexError:
                # Malformed - name: line, skip
                candidate_comments = []  # Reset comments
                continue

            # Clean task name (remove surrounding quotes if they match)
            if (task_name_raw.startswith("'") and task_name_raw.endswith("'")) or \
               (task_name_raw.startswith('"') and task_name_raw.endswith('"')):
                task_name = task_name_raw[1:-1]
            else:
                task_name = task_name_raw

            # For markdown compatibility
            task_name = task_name.replace("|", "¦")

            comment_to_assign = ""
            if candidate_comments:
                # Assign all collected candidate_comments, joined by newline
                comment_to_assign = "\n".join(candidate_comments)

            if comment_to_assign:  # Only add if there's a comment
                output_task_comments.append({
                    "task_name": task_name,
                    "task_comments": comment_to_assign
                })

            candidate_comments = []  # Reset for the next task

        elif not stripped_line:  # An empty line
            # An empty line always breaks the contiguity of comments for the next task.
            candidate_comments = []

        else:  # Any other type of line (e.g., module call, different list item)
            # These lines break the contiguity of comments leading to a task name.
            candidate_comments = []

    return output_task_comments

def get_task_line_numbers(filepath):
    with open(filepath, 'r', encoding='utf-8') as f:
        lines = f.readlines()

    tasks_lines = {}
    for idx, line in enumerate(lines):
        stripped_line = line.strip()

        if stripped_line.startswith("- name:"):
            task_name = stripped_line.replace(
                "- name:", "").split("#")[0].strip().replace("|", "¦").replace("'", "").replace('"', "")
            tasks_lines[task_name] = idx + 1

    return tasks_lines



import re
import subprocess
from pathlib import Path
from typing import Dict
from urllib.parse import urlparse, urlunparse


class GitInfoError(Exception):
    pass


class GitCommandError(GitInfoError):
    def __init__(self, message: str, stderr: str | None = None):
        super().__init__(message)
        self.stderr = stderr


class GitTimeoutError(GitInfoError):
    pass


class NotGitRepositoryError(GitInfoError):
    pass


def clean_and_standardize_url(url: str) -> str:
    processed_url = url

    scp_like_match = re.match(r"^git@([^:]+):(.*)$", processed_url)
    if scp_like_match:
        hostname = scp_like_match.group(1)
        path = scp_like_match.group(2)
        processed_url = f"ssh://git@{hostname}/{path}"

    try:
        parsed = urlparse(processed_url)
        netloc = parsed.netloc
        path = parsed.path
        force_https = False

        if "@" in netloc:
            force_https = True
            netloc_parts = netloc.rsplit('@', 1)
            if len(netloc_parts) == 2:
                netloc = netloc_parts[1]
            else:
                return ""

        if parsed.scheme in ("ssh", "git"):
            force_https = True

        final_scheme = "https" if force_https and netloc else parsed.scheme

        path = parsed.path.rstrip("/")
        if path.endswith(".git"):
            path = path[:-4]

        return urlunparse((
            final_scheme,
            netloc,
            path,
            "",
            "",
            ""
        ))
    except (ValueError, IndexError):
        return ""


def get_repo_info(path: str | Path) -> Dict[str, str]:
    dir_path = str(path)
    timeout = 5

    try:
        is_repo_check = subprocess.run(
            ["git", "-C", dir_path, "rev-parse", "--is-inside-work-tree"],
            capture_output=True, text=True, check=True, timeout=timeout
        )
        if is_repo_check.stdout.strip() != 'true':
            raise NotGitRepositoryError(f"Path is not inside a Git work tree: {dir_path}")

        raw_url = subprocess.run(
            ["git", "-C", dir_path, "config", "--get", "remote.origin.url"],
            capture_output=True, text=True, check=True, timeout=timeout
        ).stdout.strip()

        branch = subprocess.run(
            ["git", "-C", dir_path, "rev-parse", "--abbrev-ref", "HEAD"],
            capture_output=True, text=True, check=True, timeout=timeout
        ).stdout.strip()

    except subprocess.TimeoutExpired as e:
        raise GitTimeoutError(f"Git command timed out for path: {dir_path}") from e
    except subprocess.CalledProcessError as e:
        raise GitCommandError(
            f"A Git command failed for path: {dir_path}",
            stderr=e.stderr.strip()
        ) from e

    repository_url = clean_and_standardize_url(raw_url)
    hostname = urlparse(repository_url).hostname or ""

    repo_type = "default"
    if "github" in hostname:
        repo_type = "github"
    elif "gitlab" in hostname:
        repo_type = "gitlab"
    elif "gitea" in hostname:
        repo_type = "gitea"
    elif "bitbucket.org" in hostname:
        repo_type = "bitbucket"

    return {
        "repository": repository_url,
        "branch": branch,
        "repository_type": repo_type,
    }



"""Module with function for manage block and rescue code"""


def escape_pipes(text):
    """Function to escape pipes in string or list"""
    if isinstance(text, str):
        return text.replace("|", r"¦")
    if isinstance(text, list):
        return [escape_pipes(item) for item in text]
    return text  # Return the text as is if it's not a string or list.


def process_special_task_keys(task, task_type='task'):
    """Function to process tasks, including block and rescue constructs."""
    tasks = []
    known_task_params = {
        # All known Ansible task parameters.
        'action', 'any_errors_fatal', 'args', 'async', 'become', 'become_exe',
        'become_flags', 'become_method', 'become_user', 'changed_when', 'check_mode',
        'collections', 'connection', 'debugger', 'delay', 'delegate_facts', 'delegate_to',
        'diff', 'environment', 'failed_when', 'ignore_errors', 'ignore_unreachable',
        'local_action', 'loop', 'loop_control', 'module_defaults', 'name', 'no_log',
        'notify', 'poll', 'port', 'register', 'remote_user', 'retries', 'run_once',
        'tags', 'throttle', 'timeout', 'until', 'vars', 'when', 'with_', 'block',
        'rescue', 'always', 'include', 'include_tasks', 'include_role',
        'import_playbook', 'import_tasks', 'import_role', 'hosts', 'gather_facts',
        'roles', 'tasks', 'handlers', 'post_tasks', 'pre_tasks', 'strategy',
        'max_fail_percentage', 'serial', 'gather_subset', 'gather_timeout',
        'vars_files', 'vars_prompt', 'force_handlers', 'skip_tags', 'pause',
        'prompt', 'wait_for', 'wait_for_connection', 'meta', 'fact_path',
        'host_vars', 'group_vars', 'role'
    }

    for block_type in ('block', 'rescue', 'always'):
        if block_type in task:
            task_name = task.get('name', f'Unnamed_{block_type}')
            task_module = block_type
            task_when = escape_pipes(task.get('when', None))
            tasks.append({
                'name': escape_pipes(task_name),
                'module': task_module,
                'type': block_type,
                'when': task_when
            })
            for sub_task in task[block_type]:
                processed_tasks = process_special_task_keys(
                    sub_task, block_type)
                tasks.extend(processed_tasks)
            return tasks  # Exit after processing block, rescue, or always

    # Handle regular tasks
    task_name = task.get('name', 'Unnamed')
    task_when = escape_pipes(task.get('when', None))

    # Determine module name based on known task indicators or default to 'unknown'
    task_module = 'unknown'  # Default module if not found
    if 'action' in task:
        action = task['action']
        if isinstance(action, dict):
            # Module name from action dict
            task_module = list(action.keys())[0]
        else:
            task_module = action  # Module name as action string
    else:
        # Specific modules without 'action' key
        for key in ('include_tasks', 'import_tasks', 'import_playbook', 'include_role', 'import_role'):
            if key in task:
                task_module = key
                break

    # Ensure only relevant modules are shown and not general parameters like 'name' or 'when'
    if task_module == 'unknown':
        module_keys = [key for key in task.keys()
                       if key not in known_task_params and not key.startswith('with_')]
        task_module = module_keys[0] if module_keys else 'unknown'

    tasks.append({
        'name': escape_pipes(task_name),
        'module': task_module if task_module != 'unknown' else '',  # Blank if unknown
        'type': task_type,
        'when': task_when
    })
    return tasks
